#!/usr/bin/env python
# coding: utf-8

import os
import sys
sys.path.insert(0, "/home/gridsan/hwpang/Software/RMG-Py/")

from arkane.encorr.reference import ReferenceDatabase
from arkane.encorr.ae import SPECIES_LABELS, AEJob
from arkane.ess.factory import ess_factory
from arkane.modelchem import LevelOfTheory

import rmgpy.constants as constants

# load database
database = ReferenceDatabase()
database.load()

# create dictionary to map species labels to folder names from ARC's output
SPECIES_LABELS = {
    'Br2': 'Dibromine',
    'BrH': 'Hydrogen bromide',
    'CH3': 'Methyl',
    'CH3Cl': 'Chloromethane',
    'CH4': 'Methane',
    'Cl2': 'Dichlorine',
    'ClH': 'Hydrogen chloride',
    'F2': 'Difluorine',
    'FH': 'Hydrogen fluoride',
    'H2': 'Dihydrogen',
    'H2O': 'Water',
    'H2S': 'Hydrogen sulfide',
    'H3N': 'Ammonia',
    'N2': 'Dinitrogen',
    'O2': 'Dioxygen',
    'S2': 'Disulfur',
}


# In[33]:


# load experimental geometries from reference database
exp_geometries = {}
for label in SPECIES_LABELS.values():
    ref_spec = database.get_species_from_label(label)[0]
    xyz = ref_spec.reference_data['CCCBDB'].xyz_dict
    exp_geometries[label] = xyz


# In[280]:


############# USER INPUT #############


level_of_theories = {
    'B2PLYPD3': LevelOfTheory(method='B2PLYPD3-2023',
                              basis='def2tzvp',
                              software='gaussian',
                              ),

    'b3lyp': LevelOfTheory(method='b3lyp-2023',
                           basis='def2tzvp',
                           software='gaussian',
                            ),
    
    'b97d3': LevelOfTheory(method='b97d3-2023',
                           basis='def2tzvp',
                           software='gaussian',
                           ),
    
    'BP86': LevelOfTheory(method='BP86-2023',
                          basis='def2tzvp',
                          software='gaussian',
                          ),
    
    'CAM-B3LYP': LevelOfTheory(method='CAM-B3LYP-2023',
                               basis='def2tzvp',
                               software='gaussian',
                            ),

    'cbs-qb3': LevelOfTheory(method='cbsqb3-2023',
                             software='gaussian'),

    'dlpno-ccsdt-def2-TZVP': LevelOfTheory(method='dlpno-ccsd(t)-2023', basis='def2-tzvp', software='orca'),

    'dlpno-ccsdt-f12-ccpvdz': LevelOfTheory(method='dlpno-ccsd(t)-f12-2023', basis='cc-pvdz-f12', software='orca'),

    'dlpno-ccsdt-f12-ccpvtz': LevelOfTheory(method='dlpno-ccsd(t)-f12-2023', basis='cc-pvtz-f12', software='orca'),

    'g4': LevelOfTheory(method='g4-2023',
                        software='gaussian'),

    'M062X': LevelOfTheory(method='M062X-2023',  
                           basis='def2tzvp',
                           software='gaussian',
                           ),

    'PBE1PBE': LevelOfTheory(method='PBE1PBE-2023',  
                           basis='def2tzvp',
                           software='gaussian',
                           ),

    'TPSSTPSS': LevelOfTheory(method='TPSSTPSS-2023',
                              basis='def2tzvp',
                              software='gaussian',
                              ),

    'wb97mv': LevelOfTheory(method='wb97mv-2023',
                            basis='def2tzvpd',
                            software='qchem',
                            ),

    'wb97xd': LevelOfTheory(method='wb97xd-2023',
                              basis='def2tzvp',
                              software='gaussian',
                              ),              
}

arc_project_name = 'aec_bac'

# # Step 2: After running sp calculations, get AEC


# `arkane/encorr.ae.py` says:
# 
# ```
# Notes:
#     The species energies should be provided as a dictionary
#     containing the species labels as keys and their single-
#     point electronic energies in Hartree as values. The
#     energies should be calculated using the experimental
#     geometry provided for the species in the reference
#     database, and **the zero-point energy should not be included
#     in the electronic energy.**
#             
# ```
# 
# The `load_energy()` method on Arkane adapters adapter returns the energy in J/mol and does NOT include the zero-point energy. So just convert the energy from J/mol to Hartrees before fitting.

# In[282]:

for method, lot in level_of_theories.items():

    print(f"Running AEC for {method}...")
    print(f"Level of theory: {lot}")

    species_energies = {}

    if method in ['cbs-qb3', 'g4']:
        calc_directory = f"aec_calculations/{method}/calcs/Species"
        for root, dirs, files in os.walk(calc_directory):
            for file in files:
                if file.endswith("output.out"):
                    output = os.path.join(root, file)
                    print(output)
                    log = ess_factory(output)
                    energy = log.load_energy()
                    energy = energy / (constants.E_h * constants.Na)  # convert from J/mol to Hartrees
                    label = output.split("/")[-3]
                    species_energies[label] = energy
                    
    elif method == "dlpno-ccsdt-f12-ccpvdz":

        species_energies = {
            "CH3": -39.768911997620,
            "CH3Cl": -499.584136109546,
            "CH4": -40.446326748348,
            "Cl2": -919.449748451527,
            "ClH": -460.348253448604,
            "F2": -199.341426375405,
            "FH": -100.365403941669,
            "H2": -1.173236926614,
            "H2O": -76.353172147273,
            "H2S": -398.945241637250,
            "H3N": -56.488311692905,
            "N2": -109.398986853041,
            "O2": -150.163624484047,
            "S2": -795.470215490222,
        }
        
    elif method == "dlpno-ccsdt-f12-ccpvtz":

        species_energies = {
            "CH3": -39.775063989378,
            "CH3Cl": -499.608771113898,
            "CH4": -40.453254457994,
            "Cl2": -919.485516225313,
            "ClH": -460.366699031236,
            "F2": -199.379000470601,
            "FH": -100.384078117681,
            "H2": -1.173917443723,
            "H2O": -76.367718873194,
            "H2S": -398.959882039874,
            "H3N": -56.498245315894,
            "N2": -109.413149241547,
            "O2": -150.188161125483,
            "S2": -795.497492181531,
        }

    else:

        for root, dirs, files in os.walk(os.path.join("aec_calculations", method)):
            for file in files:
                if file.endswith(".log"):
                    label = file.split(".log")[0]
                    if label in SPECIES_LABELS:
                        output = os.path.join(root, file)
                        print(output)
                        log = ess_factory(output)
                        energy = log.load_energy()
                        energy = energy / (constants.E_h * constants.Na)  # convert from J/mol to Hartrees
                        species_energies[label] = energy

    print(f"{len(species_energies)} out of 16 reference species converged.")
    print(species_energies)


    species_energies = {SPECIES_LABELS[k]: v for k, v in species_energies.items()}

    ae = AEJob(species_energies=species_energies,
            level_of_theory=lot,
            )

    output_directory = os.path.join("outputs_train", f"{method}")
    os.makedirs(output_directory, exist_ok=True)

    ae.execute(output_directory=output_directory, plot=True)


