#!/usr/bin/env python
# coding: utf-8

# == modified by Oscar to not use ARC
# 
# This notebook has 3 main steps:
# - Step 1: Use ARC's API to conveniently run geometry optimizations and frequency calculations for ~400 reference species in preparation for BAC fitting
#     - Recommended to comment out some lines in `ARC/arc/job/trsh.py` to ensure that ARC will only only troubleshoot by changing the optimization algorithm and not be changing the LoT or the software. For example, if using Gaussian to calculate BACs at some LoT, it would be wise to comment out these [lines](https://github.com/ReactionMechanismGenerator/ARC/blob/master/arc/job/trsh.py#L837-L844) and these [lines](https://github.com/ReactionMechanismGenerator/ARC/blob/master/arc/job/trsh.py#L852-L870).
# - Step 2: Use Arkane to calculate H298 and add the values for the new LoT to our reference database
# - Step 3: Finally, use Arkane to perform BAC fitting
# 
# Briefly, [ARC](https://github.com/ReactionMechanismGenerator/ARC) (Automated Rate Calculator) is a software developed by the RMG team for automating electronic structure calculations, which are then used to calculate thermochemical and kinetic parameters. Installation instructions can be found on ARC's documentation [link](https://reactionmechanismgenerator.github.io/ARC/installation.html).

# In[33]:


# Add RMG-Py and ARC to Python Path (Optional)
import sys
# this is an example path if running the notebook on supercloud
# replace <USERNAME> with your username
sys.path.insert(0, "/home/gridsan/hwpang/Software/RMG-Py/")

import os
import shutil
from copy import deepcopy
import numpy as np
import subprocess

from arkane.encorr.reference import ReferenceDatabase
from arkane.ess import ess_factory
from rmgpy.molecule import Molecule

# In[35]:


import sys
sys.path.insert(0, "/home/gridsan/hwpang/Software/RDMC/")


# load the reference database
database = ReferenceDatabase()
database.load()


# # check gaussian files for normal termination before proceed

# # Step 2) Use Arkane to calculate H298 

# In[37]:


from collections import defaultdict, Counter
import os
import re

from rdkit.Chem import GetPeriodicTable

from rmgpy.quantity import ScalarQuantity
from rmgpy.statmech import HarmonicOscillator, IdealGasTranslation, LinearRotor, NonlinearRotor
from rmgpy.thermo import ThermoData

from arkane.common import symbol_by_number
from arkane.encorr.corr import get_atom_correction, assign_frequency_scale_factor
from arkane.encorr.reference import CalculatedDataEntry, ReferenceDatabase
from arkane.modelchem import LevelOfTheory, CompositeLevelOfTheory


# ## Step 2.1) Define level of theory

level_of_theories = {
    'B2PLYPD3': LevelOfTheory(method='B2PLYPD3-2023',
                              basis='def2tzvp',
                              software='gaussian',
                              ),

    'b3lyp': LevelOfTheory(method='b3lyp-2023',
                           basis='def2tzvp',
                           software='gaussian',
                            ),
    
    'b97d3': LevelOfTheory(method='b97d3-2023',
                           basis='def2tzvp',
                           software='gaussian',
                           ),
    
    'BP86': LevelOfTheory(method='BP86-2023',
                          basis='def2tzvp',
                          software='gaussian',
                          ),
    
    'CAM-B3LYP': LevelOfTheory(method='CAM-B3LYP-2023',
                               basis='def2tzvp',
                               software='gaussian',
                            ),

    'cbs-qb3': LevelOfTheory(method='cbsqb3-2023',
                             software='gaussian'),

    'dlpno-ccsdt-def2-TZVP': CompositeLevelOfTheory(freq=LevelOfTheory(method='wb97xd-2023', basis='def2tzvp', software='gaussian'),
                                                     energy=LevelOfTheory(method='dlpno-ccsd(t)-2023', basis='def2-tzvp', software='orca')),

    'dlpno-ccsdt-f12-ccpvdz': CompositeLevelOfTheory(freq=LevelOfTheory(method='wb97mv-2023', basis='def2tzvpd', software='qchem'),
                                                     energy=LevelOfTheory(method='dlpno-ccsd(t)-f12-2023', basis='cc-pvdz-f12', software='orca')),

    'dlpno-ccsdt-f12-ccpvtz': CompositeLevelOfTheory(freq=LevelOfTheory(method='wb97mv-2023', basis='def2tzvpd', software='qchem'),
                                                     energy=LevelOfTheory(method='dlpno-ccsd(t)-f12-2023', basis='cc-pvtz-f12', software='orca')),

    'g4': LevelOfTheory(method='g4-2023',
                        software='gaussian'),

    'M062X': LevelOfTheory(method='M062X-2023',  
                           basis='def2tzvp',
                           software='gaussian',
                           ),

    'PBE1PBE': LevelOfTheory(method='PBE1PBE-2023',
                             basis='def2tzvp',
                             software='gaussian',
                             ),

    'TPSSTPSS': LevelOfTheory(method='TPSSTPSS-2023',
                              basis='def2tzvp',
                              software='gaussian',
                              ),

    'wb97mv': LevelOfTheory(method='wb97mv-2023',
                            basis='def2tzvpd',
                            software='qchem',
                            ),

    'wb97xd': LevelOfTheory(method='wb97xd-2023',
                            basis='def2tzvp',
                            software='gaussian',
                            ),              
}

# task_id = int(sys.argv[1])
# level_of_theories = {k: v for i, (k, v) in enumerate(level_of_theories.items()) if i == task_id}

freq_scale_factors = {
    "wb97xd": 0.9914,
    "b3lyp": 1.0044,
    "M062X": 0.9871,
    "PBE1PBE": 0.9944,
    "TPSSTPSS": 1.0194,
    "BP86": 1.0337,
    "b97d3": 0.988,
    "CAM-B3LYP": 0.990,
    "B2PLYPD3": 0.995,
    "wb97mv": 1.002,
    "dlpno-ccsdt-f12-ccpvdz": 1.002,
    "dlpno-ccsdt-f12-ccpvtz": 1.002,
    "dlpno-ccsdt-def2-TZVP": 0.9914,
    "g4": 1.0000,
    "cbs-qb3": 1.00386,
}

def log_to_enthalpy(method, freq_lot, energy_lot, freq_log, energy_log=None, energy_dict=None, temp=298.15):
    print(freq_log)

    freq_log = ess_factory(freq_log)
    
    conformer, _ = freq_log.load_conformer()
    
    # Perform quick checks
    assert conformer.spin_multiplicity > 0
    assert any(isinstance(mode, IdealGasTranslation) for mode in conformer.modes)
    assert any(isinstance(mode, (LinearRotor, NonlinearRotor)) for mode in conformer.modes)
    assert any(isinstance(mode, HarmonicOscillator) for mode in conformer.modes)
    
    coords, nums, masses = freq_log.load_geometry()
    assert len(nums) > 1
    
    atoms = Counter([symbol_by_number[int(n)] for n in nums]) 
    conformer.coordinates = (coords, 'angstroms')
    conformer.number = nums
    conformer.mass = (masses, 'amu')
    
    freq_scale_factor = freq_scale_factors[method]
    frequencies = conformer.modes[2].frequencies.value_si
    for mode in conformer.modes:
        if isinstance(mode, HarmonicOscillator):
            mode.frequencies = (frequencies * freq_scale_factor, "cm^-1")
    if freq_scale_factor == 1:
        print('WARNING: Frequency scale factor is 1')
    zpe_scale_factor = freq_scale_factor / 1.014
    
    # get electronic energy (choose one option)
    # option 1: read from the log
    if energy_log is not None:
        energy_log = ess_factory(energy_log)
        energy = energy_log.load_energy(zpe_scale_factor=zpe_scale_factor)  # J/mol

    # option 2: just read from a dictionary provided 
#     energy = energy_dict[int(label)] * 627.5094740631 * 4184
    
    # add ZPE
    energy += freq_log.load_zero_point_energy() * zpe_scale_factor if len(nums) > 1 else 0  # J/mol
    
    # add AECs
#     print(get_atom_correction(energy_lot, atoms) )
    energy += get_atom_correction(energy_lot, atoms)  # J/mol
    conformer.E0 = (energy / 4184, 'kcal/mol')
    
    return ScalarQuantity((conformer.get_enthalpy(temp) + conformer.E0.value_si) / 4184, 'kcal/mol')

periodic_table = GetPeriodicTable()

def log_to_xyz_dict(log):
    log = ess_factory(log)
    coords, nums, _ = log.load_geometry()
    syms = [symbol_by_number[int(n)] for n in nums]
    return {
        'coords': coords,
        'isotopes': [periodic_table.GetMostCommonIsotope(s) for s in syms],
        'symbols': syms
    }


# ## Step 2.3) Match all of the reference species to log files

# In[56]:


re_int = re.compile(r'\d+')
ref_spcs = {spc.index: spc for spc in database.reference_sets['main']}

# Each log file must start with the index followed by an underscore
def get_logs(log_dir):
    if log_dir is None:
        return defaultdict(lambda: None)
    
    logs = {}
    for root, dirs, files in os.walk(log_dir):
        for file in files:
            if file.endswith(".log") or file.endswith(".out"):
                matched = re_int.search(file)
                if matched is None:
                    continue
                spc_ind = int(matched.group())
                log_path = os.path.join(root, file)
                logs[spc_ind] = log_path
    
    num_missing = 0
    for missing_idx in ref_spcs.keys() - logs.keys():
        print(f'{ref_spcs[missing_idx]} is missing a log file in {log_dir}')
        num_missing += 1
    print(f'A total of {num_missing} reference species are missing a log file!\n')    
    
    return logs


def check_job_status(self):
    
    for line in reversed(self): 

        if 'Normal termination' in line:
            return True
        else:
            return False
        
def load_freq(self):
    """
    Return the negative frequency from a transition state frequency
    calculation in cm^-1.
    """
    frequencies = []
    with open(self, 'r') as f:
        line = f.readline()
        while line != '':
            # Read vibrational frequencies
            if 'Frequencies --' in line:
                frequencies.extend(line.split()[2:])
            line = f.readline()

    frequencies = [float(freq) for freq in frequencies]
    frequencies.sort()
    
    return frequencies

cwd = os.getcwd()
cwd

output_dir = os.path.join(cwd, "outputs")

for method, lot in level_of_theories.items():

    output_directory = os.path.join(output_dir, f"{method}")
    os.makedirs(output_directory, exist_ok=True)

    arkane_input_path = os.path.join(output_directory, "bac_arkane.py")

    if os.path.exists(arkane_input_path):
        print(f"Arkane input file already exists in {arkane_input_path}.")
        continue

    print(lot)

    if isinstance(lot, CompositeLevelOfTheory):
        freq_lot = lot.freq
        energy_lot = lot.energy
    else:
        freq_lot = lot
        energy_lot = lot

    if method in ['dlpno-ccsdt-f12-ccpvdz', 'dlpno-ccsdt-f12-ccpvtz']:
        # for dlpno methods, energy log dir is not the same as freq log dir
        freq_log_dir = os.path.join(cwd, 'bac_calculations', 'wb97mv')
        energy_log_dir = os.path.join(cwd, 'bac_calculations', method)
    elif method in ['dlpno-ccsdt-def2-TZVP']:
        freq_log_dir = os.path.join(cwd, 'bac_calculations', 'wb97xd')
        energy_log_dir = os.path.join(cwd, 'bac_calculations', method)
    elif method in ['cbs-qb3', 'g4']:
        freq_log_dir = os.path.join(cwd, 'bac_calculations', method, 'results')
        energy_log_dir = freq_log_dir
    else:
        freq_log_dir = os.path.join(cwd, 'bac_calculations', method)
        energy_log_dir = freq_log_dir

    print("freq_log_dir: ", freq_log_dir)
    print("energy_log_dir: ", energy_log_dir)

    # ## Step 2.2) Define a function for using the Arkane API to calculate thermo
        
    freq_logs = get_logs(freq_log_dir)
    energy_logs = get_logs(energy_log_dir)

    # sanity check
    assert len(freq_logs) > 400
    assert len(energy_logs) > 400

    # assert len(energy_logs) == len(freq_logs)


    # ## Step 2.4) Compute thermo and update values for the reference species

    logs_iter = freq_logs if energy_log_dir is None else energy_logs
            
    # success_id = [] # exclude failed qm jobs, provided by user
    # for spc_id, log_path in logs_iter.items():
    #     with open(log_path, "r") as f:
    #         lines = f.readlines()
    #         normal_termination = check_job_status(lines) # not able to check for qchem and orca...
    #     frequencies = load_freq(log_path)
    #     frequencies = np.array(frequencies)
    #     if normal_termination and np.all(frequencies > 0):
    #         success_id.append(spc_id)

    # print("success_id: ", success_id)

    success_id = list(logs_iter.keys())

    for k in list(logs_iter.keys()):
        if k not in success_id:
            del logs_iter[k]


    assert set(list(logs_iter.keys())) == set(success_id)


    # In[63]:


    exclude = [] # additional jobs to exclude, provided by user


    # In[ ]:





    # In[64]:

    print(logs_iter.keys())

    for idx in logs_iter.keys():
        
        # ignore not successful sp
        if idx not in success_id:
            continue
        
        # Ignore extra logs that we don't have ref species for
        if idx not in ref_spcs:
            continue
            
        # additional jobs to exclude    
        if idx in exclude:
            continue

        try:
            # modify the following line, depends on whether energy is read from file or from dict
            hf298 = log_to_enthalpy(method, freq_lot, energy_lot, freq_logs[idx], energy_log=energy_logs[idx]) 
            
            # Update thermo if already in calculated data
            if lot in ref_spcs[idx].calculated_data:  
                ref_spcs[idx].calculated_data[lot].thermo_data.H298 = hf298
            # Make new calculated data entry otherwise
            else:
                ref_spcs[idx].calculated_data[lot] = CalculatedDataEntry(ThermoData(H298=hf298), xyz_dict=log_to_xyz_dict(energy_logs[idx]))
        except Exception as e:
            print(f'Error for {energy_logs[idx]}: {e}')
            continue


    # ## Step 2.5) Save the new enthalpy values to RMG-database
    # 

    # In[65]:


    database.save()


    # # Step 3) Create Arkane input file for BAC fitting

    # In[66]:



    # In[67]:


    bac_input = f"""lot = {lot.__repr__()}

kwargs = {{
    'crossval_n_folds': -1,
    'weighted': True,
}}

# Petersson-type
bac(
    level_of_theory=lot,
    bac_type='p',  # Petersson
    **kwargs
)

# Melius-type
bac(
    level_of_theory=lot,
    bac_type='m',  # Melius
    **kwargs
)
"""

    with open(arkane_input_path, "w") as f:
        f.write(bac_input)

    print(f"Arkane input file written to {arkane_input_path}.")


    # In[70]:



    # In[ ]:


    # subprocess.run(f"python /home/gridsan/hwpang/Software/RMG-Py/Arkane.py bac_arkane.py",
    #             shell=True,
    #             cwd=output_directory,
    #             )


# Then do `python RMG-Py/Arkane.py bac_arkane.py`

print("Done!")